# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of python's std lib"""
import collections
from copy import deepcopy
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    Dict,
    Iterable,
    List,
    Mapping,
    Optional,
    Set,
    Tuple,
    Type,
    Union,
    _GenericAlias,
)

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.type_check_error import (
    HashableError,
    KeyValueError,
    TupleArgLenError,
    TypeCheckError,
    UnRepresentedTypeError,
    WrongTypeError,
)
from adorn.unit.simple import Simple


class Python(Simple):
    """:class:`~adorn.unit.unit.Unit` that holds types in python's std lib."""

    _registry = collections.defaultdict(dict)

    def get(self, key: Any, orchestrator: "Orchestrator") -> Optional[type]:
        """Return the relevant python std lib type

        Args:
            key (Any): the type that may be part of python's
                std lib
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[type]: if ``key`` is part of python's std lib,
                then the type that is meant to represent the python
                std lib type will be returned, otherwise ``None``
        """
        if isclass(key) or isinstance(key, _GenericAlias):
            return self._get(key, orchestrator)
        return None

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the type is part of python's std lib.

        Args:
            obj (Any): the type that may be part of python's
                std lib
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is part of python's std lib
        """
        if not (isclass(obj) or isinstance(obj, _GenericAlias)):
            return False
        return any(self.relevant_types(obj, orchestrator))


class BuiltIn(Python):
    """Representation of python's built-in types"""

    base_type: type
    """the built-in python type the subclass is supposed to represent"""

    hashable: bool = True
    """python's built-in types are hashable"""

    unacceptable_subtype: Set[type] = {}
    """types that do not map to the
       given python built-in type.  This exists to prevent ``bool``
       from being confused as an ``int``."""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is part of python's built-in types.

        Args:
            obj (Type): the type that may be part of python's
                built-in types
            orchestrator (Orchestrator): container of all types, which
                is unused, since there are no nested types

        Returns:
            bool: if ``True``, the type is part of python's std lib
        """
        return (not isinstance(obj, _GenericAlias)) and (
            (issubclass(obj, cls.base_type) or (issubclass(obj, cls)))
            and obj not in cls.unacceptable_subtype
        )

    @classmethod
    def _from_obj(cls, target_cls: type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Instantiate an instance of the python built-in type

        Args:
            target_cls (type): The python built-in class, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is unused
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of the python built-in type, as specified
                by ``target_cls``
        """
        return cls.base_type(obj)

    @classmethod
    def additional_checks(
        cls, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Type checks specific to a given type

        Args:
            orchestrator (Orchestrator): container of all types,
                typically unused
            obj (Any): an instance that will have additional
                type checking performed against it

        Returns:
            Optional[TypeCheckError]: either a ``TypeCheckError`` containing
                the reason why the ``obj`` didn't pass the additional checks,
                otherwise ``None``
        """
        return None

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check if an instance could be converted to the desired python built-in type

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` was not of the appropriate type

        Args:
            target_cls (Type): The python built-in class, ``obj``
                will be checked against
            orchestrator (Orchestrator): container of all types,
                which is unused
            obj (Any): an instance that will be checked against
                the type ``target_cls``

        Returns:
            Optional[TypeCheckError]: either a ``TypeCheckError`` containing
                the reason why the ``obj`` didn't type check, otherwise
                ``None``
        """
        if not isinstance(obj, cls.base_type):
            return WrongTypeError(target_cls=cls.base_type, obj=obj)
        return cls.additional_checks(orchestrator=orchestrator, obj=obj)


class Wrapper(Python):
    """Base class for python wrapper types."""

    @staticmethod
    def is_hashable(
        target_cls: Type,
        target_arg: Type,
        orchestrator: "Orchestrator",
        obj: Iterable[Any],
    ) -> Optional[TypeCheckError]:
        """Check if a given type may be hashed.

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                the nested argument, did not map to any of the types contained in
                the :class:`~adorn.orchestrator.orchestrator.Orchestrator`
            - :class:`~adorn.exception.type_check_error.HashableError`:
                the nested argument, could not be hashed

        Args:
            target_cls (Type): fully instantiated wrapper type
            target_arg (Type): nested type, that is required to be hashable
            orchestrator (Orchestrator): container of all types, used to recurse
            obj (Iterable[Any]): the instance, pre conversion, used to enrich
                error messages, if they occur

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue with the nested type, otherwise ``None``
        """
        if not orchestrator.contains(target_arg):
            return UnRepresentedTypeError(
                target_cls=target_cls, unit=orchestrator, obj=obj
            )
        unit = orchestrator.get(target_arg).get(target_arg, orchestrator)
        if not unit.hashable:
            return HashableError(target_cls=target_cls, unit=unit, obj=obj)


@Python.register("int")
class Int(BuiltIn):
    """Representation of python's ``int`` type"""

    base_type = int
    unacceptable_subtype: Set[type] = {bool}

    @classmethod
    def additional_checks(
        cls, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Type checks specific to ``int``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` was of type ``bool``, not ``int``

        Args:
            orchestrator (Orchestrator): container of all types,
                unused
            obj (Any): an instance that will have additional
                type checking performed against it

        Returns:
            Optional[TypeCheckError]: either a ``TypeCheckError`` containing
                the reason why the ``obj`` didn't pass the additional checks,
                otherwise ``None``
        """
        if type(obj) == bool:  # noqa: E721
            return WrongTypeError(int, obj)


@Python.register("str")
class Str(BuiltIn):
    """Representation of python's ``str`` type"""

    base_type = str


@Python.register("bool")
class Bool(BuiltIn):
    """Representation of python's ``bool`` type"""

    base_type = bool


@Python.register("float")
class Float(BuiltIn):
    """Representation of python's ``float`` type"""

    base_type = float


@Python.register("list")
class Rterable(Wrapper):
    """Representation of python's ``Iterable`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Iterable``

        Args:
            obj (Type): the type that may be ``Iterable``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested type

        Returns:
            bool: if ``True``, the type is ``Iterable``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return (
            origin in {collections.abc.Iterable, Iterable, List, list}
            and len(args) == 1
            and orchestrator.contains(args[0])
        )

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Any:
        """Instantiate an instance of ``Iterable``

        Args:
            target_cls (type): The fully instantiated ``Iterable`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the elements of the ``Iterable``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Iterable``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return [orchestrator.from_obj(args[0], i) for i in obj]

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Optional[TypeCheckError]:
        """Check if an instance could be converted to ``Iterable``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to the nested type the ``Iterable`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to an ``Iterable``

        Args:
            target_cls (Type): The fully instantiated ``Iterable`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Iterable``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_arg = getattr(target_cls, "__args__", [])[0]
        if (not isinstance(obj, set)) and any(
            isinstance(obj, i) for i in [collections.abc.Iterable, Iterable, List, list]
        ):
            list_checks = (orchestrator.type_check(target_arg, i) for i in obj)
            relevant_checks = [
                (en, i) for en, i in enumerate(list_checks) if i is not None
            ]
            if relevant_checks:
                key_values = dict(relevant_checks)
                return KeyValueError(
                    target_cls=target_cls, key_values=key_values, obj=obj
                )
            return None
        return WrongTypeError(target_cls=target_cls, obj=obj)


@Python.register("set")
class Ret(Wrapper):
    """Represents Python's ``Set`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Set``

        Args:
            obj (Type): the type that may be ``Set``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested type

        Returns:
            bool: if ``True``, the type is ``Set``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return (
            origin in {Set, set} and len(args) == 1 and orchestrator.contains(args[0])
        )

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Set[Any]:
        """Instantiate an instance of ``Set``

        Args:
            target_cls (type): The fully instantiated ``Set`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the elements of the ``Set``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Set``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return {orchestrator.from_obj(args[0], i) for i in obj}

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Iterable[Any]
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Set``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                the nested argument of ``Set``, did not map to any of the types
                contained in the :class:`~adorn.orchestrator.orchestrator.Orchestrator`
            - :class:`~adorn.exception.type_check_error.HashableError`:
                the nested argument of ``Set`` wasn't hashable
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to the nested type the ``Set`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to a ``Set``

        Args:
            target_cls (Type): The fully instantiated ``Set`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Set``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_arg = getattr(target_cls, "__args__", [None])[0]
        if (target_arg is not None) and any(
            isinstance(obj, i) for i in [collections.abc.Iterable, Iterable, List, list]
        ):
            hashable_error = cls.is_hashable(
                target_cls=target_cls,
                target_arg=target_arg,
                orchestrator=orchestrator,
                obj=obj,
            )
            if hashable_error is not None:
                return hashable_error
            list_checks = (orchestrator.type_check(target_arg, i) for i in obj)
            relevant_checks = [i for i in list_checks if i is not None]
            if relevant_checks:
                key_values = dict(enumerate(relevant_checks))
                return KeyValueError(
                    target_cls=target_cls, key_values=key_values, obj=obj
                )
            return None
        return WrongTypeError(target_cls=target_cls, obj=obj)


@Python.register("tuple")
class Ruple(Wrapper):
    """Represents Python's ``Tuple`` type

    .. note::

        This representation of ``Tuple`` does not handle tuples of
        arbitrary length.  The full length of the ``Tuple`` must
        be specified in the type signature at instantiation time.
    """

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Tuple``

        Args:
            obj (Type): the type that may be ``Tuple``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is ``Tuple``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return origin in {Tuple, tuple} and all(orchestrator.contains(i) for i in args)

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Tuple[Any, ...]
    ) -> Any:
        """Instantiate an instance of ``Tuple``

        Args:
            target_cls (type): The fully instantiated ``Tuple`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the elements of the ``Tuple``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Tuple``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return tuple(
            orchestrator.from_obj(i, j) for i, j in zip(args, obj)  # noqa: B905
        )

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Tuple[Any, ...]
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Tuple``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to the nested type the ``Tuple`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to a ``Tuple``

        Args:
            target_cls (Type): The fully instantiated ``Tuple`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Tuple``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_args = getattr(target_cls, "__args__", [])
        if (
            not isinstance(obj, dict)
            and (target_args != [])
            and (not isinstance(obj, _GenericAlias))
            and isinstance(obj, Iterable)
        ):
            if len(target_args) != len(obj):
                return TupleArgLenError(target_cls, obj)
            list_checks = [
                orchestrator.type_check(i, j)
                for i, j in zip(target_args, obj)  # noqa: B905
            ]
            relevant_checks = [
                (en, i) for en, i in enumerate(list_checks) if i is not None
            ]
            if relevant_checks:
                key_values = dict(relevant_checks)
                return KeyValueError(
                    target_cls=target_cls, key_values=key_values, obj=obj
                )
            return None
        return WrongTypeError(target_cls=target_cls, obj=obj)


@Python.register("union")
class Rnion(Wrapper):
    """Represents Python's ``Union`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Union``

        Args:
            obj (Type): the type that may be ``Union``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is ``Tuple``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return origin == Union and all(orchestrator.contains(i) for i in args)

    @classmethod
    def _from_obj(cls, target_cls: type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Instantiate an instance of ``Union``

        Args:
            target_cls (type): The fully instantiated ``Union`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate an element of the ``Union``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Union``, as specified
                by ``target_cls``

        Raises:
            KeyValueError: couldn't convert ``obj`` to any of the
                types specified in ``Union``
        """
        args = getattr(target_cls, "__args__", [])
        failed_args = []
        for arg in args:
            # do we really need to copy
            obj_prime = deepcopy(obj)
            output = orchestrator.type_check(arg, obj_prime)
            if output is None:
                return orchestrator.from_obj(arg, obj)
            else:
                failed_args.append((arg, output))

        key_values = {k: v for k, v in failed_args}
        raise KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Union``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't be converted
                to a type specified in the ``Union``

        Args:
            target_cls (Type): The fully instantiated ``Union`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Union``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_args = getattr(target_cls, "__args__", [])
        list_checks = [(i, orchestrator.type_check(i, obj)) for i in target_args]
        if all(i is not None for _, i in list_checks):
            key_values = {k: v for k, v in list_checks}
            return KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)
        return None


@Python.register("dict")
class Rict(Wrapper):
    """Represents Python's ``Dict`` type"""

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the type is ``Dict``

        Args:
            obj (Type): the type that may be ``Dict``
            orchestrator (Orchestrator): container of all types,
                used to recurse nested types

        Returns:
            bool: if ``True``, the type is ``Dict``
        """
        origin = getattr(obj, "__origin__", None)
        args = getattr(obj, "__args__", [])
        return (
            origin in {collections.abc.Mapping, Mapping, Dict, dict}
            and len(args) == 2
            and all(orchestrator.contains(i) for i in args)
        )

    @classmethod
    def _from_obj(
        cls, target_cls: type, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Dict[Any, Any]:
        """Instantiate an instance of ``Dict``

        Args:
            target_cls (type): The fully instantiated ``Dict`` type, an
                instance will be generated of
            orchestrator (Orchestrator): container of all types,
                which is used to instantiate the keys and values
                of the ``Dict``
            obj (Any): an instance that will be converted to an
                instance of type ``target_cls``

        Returns:
            Any: an instance of ``Dict``, as specified
                by ``target_cls``
        """
        args = getattr(target_cls, "__args__", [])
        return {
            orchestrator.from_obj(args[0], k): orchestrator.from_obj(args[1], v)
            for k, v in obj.items()
        }

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[str]:
        """Check if an instance could be converted to ``Dict``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                the key type of ``Dict``, did not map to any of the types
                contained in the :class:`~adorn.orchestrator.orchestrator.Orchestrator`
            - :class:`~adorn.exception.type_check_error.HashableError`:
                the key type of ``Dict`` wasn't hashable
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                ``obj`` contained elements that couldn't
                be converted to the key or value type the ``Dict`` wraps
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``obj`` couldn't be converted to a ``Dict``

        Args:
            target_cls (Type): The fully instantiated ``Dict`` type, an
                instance will be checked against
            orchestrator (Orchestrator): container of all types,
                which is used to check the elements of the ``Dict``
            obj (Any): an instance that will be checked against the
                type ``target_cls``

        Returns:
            Optional[TypeCheckError]: Reason why ``obj`` can't be converted
                to an instance of ``target_cls``, otherwise ``None``
        """
        target_args = getattr(target_cls, "__args__", [])
        if all(
            not isinstance(obj, i)
            for i in [collections.abc.Mapping, Mapping, Dict, dict]
        ):
            return WrongTypeError(target_cls=target_cls, obj=obj)
        hashable_error = cls.is_hashable(
            target_cls=target_cls,
            target_arg=target_args[0],
            orchestrator=orchestrator,
            obj=obj,
        )
        if hashable_error is not None:
            return hashable_error
        keys = []
        values = []
        for k, v in obj.items():
            # check key
            k_tc = orchestrator.type_check(target_args[0], k)
            if k_tc is not None:
                keys.append(k_tc)
            # check value
            v_tc = orchestrator.type_check(target_args[1], v)
            if v_tc:
                values.append(v_tc)

        if keys or values:
            # key(s) and/or value(s) didn't pass type_check
            key_values = dict()
            for en, (prefix, lst) in enumerate([("key", keys), ("value", values)]):
                for en_prime, i in enumerate(lst):
                    key_values[f"{prefix}_{target_args[en]}_{en_prime}"] = i
            return KeyValueError(target_cls=target_cls, key_values=key_values, obj=obj)
        return None



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of ``dataclasses.dataclass``"""
from dataclasses import is_dataclass
from inspect import isclass
from typing import TYPE_CHECKING, Any, Callable, ClassVar, Dict, Optional, Type, TypeVar

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.constructor import Constructor
from adorn.exception.type_check_error import (
    ParamError,
    TypeCheckError,
    UnRepresentedTypeError,
)
from adorn.params import Params
from adorn.unit.unit import Unit

_D = TypeVar("_D")


class DataClass(Unit):
    """A wrapper around :meth:`~dataclasses.dataclass`

    .. code-block:: python

        from dataclasses import dataclass

        from adorn.orchestrator.base import Base
        from adorn.params import Params
        from adorn.unit.dataclass import DataClass

        class Example(DataClass):
            _registry = dict()

        @Example.register()
        @dataclass
        class A:
            a: int
            b: bool

        base = Base([Example()])
        assert base.from_obj(A, Params({"a": 1, "b": False})) == A(1, False)

    """  # noqa: RST304

    _registry: ClassVar[Dict[type, type]]

    @classmethod
    def register(cls) -> Callable[[Type[_D]], Type[_D]]:  # pragma: no cover
        """Add :meth:`~dataclasses.dataclass` to the :attr:`~adorn.unit.anum.Anum._registry`

        Returns:
            Callable[[Type[_D]], Type[_D]]: funcion that adds the decorated
                :meth:`~dataclasses.dataclass` to the ``_registry``
        """  # noqa: B950, D202, RST304

        def add_members(subclass: Type[_D]) -> Type[_D]:
            cls._registry[subclass] = subclass
            return subclass

        return add_members

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional[type]:  # pragma: no cover
        """Return the relevant ``Anum`` for the `key`

        Args:
            key (Any): the Type that may be a subclass of the ``Unit``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            Optional[type]: if ``key`` is an ``Anum`` then return it,
                otherwise ``None``
        """
        if self.contains(key, orchestrator):
            return key

    def contains(
        self, obj: Any, orchestrator: "Orchestrator"
    ) -> bool:  # pragma: no cover
        """Check if `obj` is registered to ``DataClass``

        Args:
            obj (Any): the Type that may be a subclass of ``Anum``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            bool: if ``True`` then ``obj`` is represented by ``Anum``,
                otherwise ``False``
        """
        return isclass(obj) and is_dataclass(obj) and obj in self._registry

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:  # pragma: no cover
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.ParamError`:
                ``obj`` was not of type :class:`~adorn.params.Params`

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, used to check the
                arguments of the constructor for the given ``dataclass``
            obj (Any): a :class:`~adorn.params.Params`, that will be checked to
                see if it can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(target_cls, orchestrator):
            return UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)

        if not isinstance(obj, Params):
            return ParamError(target_cls=target_cls, obj=obj)

        return orchestrator.type_check(Constructor(target_cls), obj)

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> _D:
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, used to instantiate
                the arguments for the constructor of the given ``dataclass``
            obj (Any): an instance, that will be converted to an
                instance of ``target_cls``

        Returns:
            _D: an instantiated :meth:`~dataclasses.dataclass`
        """
        return orchestrator.from_obj(Constructor(target_cls), obj)



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Different ways to handle instances of :class:`~adorn.data.constructor.Constructor`"""
import inspect
from collections import defaultdict
from typing import TYPE_CHECKING, Any, Callable, Dict, Optional, Set, Type

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.constructor import Constructor
from adorn.data.parameter import Parameter
from adorn.exception.type_check_error import (
    KeyValueDiffError,
    KeyValueError,
    ParameterOrderError,
    TypeCheckError,
    WrongTypeError,
)
from adorn.unit.simple import Simple


class ConstructorValue(Simple):
    """:class:`~adorn.unit.unit.Unit` that contains all the ways to handle :class:`~adorn.data.constructor.Constructor` instances."""  # noqa: B950

    _registry = defaultdict(dict)

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is an instance of :class:`~adorn.data.constructor.Constructor`

        Args:
            obj (Any): potentially an instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): this is unused

        Returns:
            bool: if ``True``, ``obj`` is an instance of
                :class:`~adorn.data.constructor.Constructor`
        """  # noqa: B950
        return isinstance(obj, Constructor)

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional["ConstructorValue"]:
        """Get the relevant class to handle the given :class:`~adorn.data.constructor.Constructor` instance.

        The :attr:`~adorn.unit.complex.Complex.constructor_type` attached to the
        :attr:`~adorn.data.constructor.Constructor.subclass` is used to determine which
        subclass of ``ConstructorValue`` to use.  If
        :attr:`~adorn.unit.complex.Complex.constructor_type` is ``None`` then
        :class:`~adorn.unit.constructor_value.BaseConstructorValue` will be returned

        Args:
            key (Any): potentially an instance of
                :class:`~adorn.data.constructor.Constructor``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[ConstructorValue]: if ``key`` is an instance of
                :class:`~adorn.data.constructor.Constructor` then a subclass of
                ``ConstructorValue`` will be returned, otherwise ``None``
        """  # noqa: B950, RST304
        if not self.contains(key, orchestrator):
            return None
        return self._registry[ConstructorValue].get(key.constructor_type or "base")

    def from_obj(
        self, target_cls: Constructor, orchestrator: "Orchestrator", obj: Any
    ) -> Any:
        """Instantiate the arguments and an instance of :attr:`~adorn.data.constructor.Constructor.subclass`

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Any: An instance of :attr:`~adorn.data.constructor.Constructor.subclass`
        """  # noqa: B950, RST304
        constructor_value = self.get(target_cls, orchestrator)
        return constructor_value._from_obj(target_cls, orchestrator, obj)

    def type_check(
        self, target_cls: Constructor, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check the arguments for :attr:`~adorn.data.constructor.Constructor.subclass`

        Exceptions:
            - :class:`~adorn.exception.type_check_error.WrongTypeError`:
                ``target_cls`` was not an instance of
                :class:`~adorn.data.constructor.Constructor`


        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then either
                ``target_cls`` was not an instance of
                :class:`~adorn.data.constructor.Constructor`
                or there are issues which would prevent creating the arguments
                for :attr:`~adorn.data.constructor.Constructor.subclass`,
                otherwise ``None``
        """  # noqa: RST304
        if not self.contains(target_cls, orchestrator):
            return WrongTypeError(Constructor, target_cls)

        constructor_value = self.get(target_cls, orchestrator)
        return constructor_value._type_check(target_cls, orchestrator, obj)

    @staticmethod
    def get_kv(
        main: Dict[str, Any],
        alter_value: Callable[..., type],
        sub: Dict[str, Any],
        missing: Set[str],
        check: Callable[..., bool] = lambda _: False,
    ) -> Optional[Dict[str, Type]]:
        """Generate the difference in keys between ``main`` and ``sub``.

        Args:
            main (Dict[str, Any]): a dict, whose keys are considered necessary
            alter_value (Callable[..., type]): an alteration performed to a value
                of ``main`` before it is added to the returned result
            sub (Dict[str, Any]): a dict which is checked that it contains the
                keys in ``main``
            missing (Set[str]): a collection of keys to be ignored in ``main``
            check (Callable[..., bool]): an additional check on the keys of ``main``,
                which if ``True`` will prevent the key from being checked

        Returns:
            Optional[Dict[str, Type]]: the keys and associated types that were missing
                in ``sub``.  If no keys were missing then ``None``
        """
        extras = dict()
        for k in main.keys():
            if k in missing or check(main[k]):
                continue
            if k not in sub:
                extras[k] = alter_value(main[k])
        if extras:
            return extras

    @staticmethod
    def parameter_consistency(
        target_cls: Constructor, obj: Dict[Any, Any]
    ) -> Optional[KeyValueDiffError]:
        """Check that ``obj`` has all the necessary arguments, and nothing more.

        Args:
            target_cls (Constructor): contains the parameters required by
                :attr:`~adorn.data.constructor.Constructor.subclass`
            obj (Dict[Any, Any]): collection of potential arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[KeyValueDiffError]: ``None`` if ``obj`` contained the correct
                set of keys, otherwise ``KeyValueDiffError`` is returned which specifies
                the missing and/or additional arguments in ``obj``
        """  # noqa: RST304
        parameter_kv = ConstructorValue.get_kv(
            target_cls.parameters,
            lambda i: i.annotation,
            obj,
            {"self"},
            lambda i: i.default != inspect._empty,
        )
        obj_kv = ConstructorValue.get_kv(obj, type, target_cls.parameters, {"type"})

        if parameter_kv or obj_kv:
            return KeyValueDiffError(
                target_cls=target_cls.subclass,
                parameter_kv=parameter_kv,
                obj_kv=obj_kv,
                obj=obj,
            )
        return None

    @staticmethod
    def dict_type_check(
        target_cls: Constructor,
        orchestrator: "Orchestrator",
        obj: Dict[Any, Any],
    ) -> Optional[KeyValueError]:
        """Type check all the arguments in ``obj``

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`, which
                specifies the parameters for the constructor
            orchestrator (Orchestrator): container of all types, used
                to check the arguments
            obj (Any): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[KeyValueError]: if all the args type check then ``None``,
                otherwise ``KeyValueError``, which specifies the args that
                failed and why
        """  # noqa: RST304
        exception_dict = dict()
        for k in target_cls.parameter_order or obj.keys():
            if k in obj and (k not in {"type"}):
                parameter = Parameter(
                    target_cls.parameters[k].annotation, target_cls, obj, k
                )
                output = orchestrator.type_check(parameter, obj[k])
                if output is not None:
                    exception_dict[k] = output

        if exception_dict:
            return KeyValueError(
                target_cls=target_cls.subclass, key_values=exception_dict, obj=obj
            )
        return None

    @staticmethod
    def check_parameter_order(target_cls: Constructor) -> Optional[ParameterOrderError]:
        """Ensure :attr:`~adorn.unit.complex.Complex.parameter_order` specifies all the parameters.

        Args:
            target_cls (Constructor): :attr:`~adorn.data.constructor.Constructor.subclass`
                specifies a :attr:`~adorn.unit.complex.Complex.parameter_order`

        Returns:
            Optional[ParameterOrderError]: if :attr:`~adorn.unit.complex.Complex.parameter_order`
                is complete then ``None``, otherwise ``ParameterOrderError``, which specifies
                the additional and/or missing parameters in
                :attr:`~adorn.unit.complex.Complex.parameter_order`
        """  # noqa: B950, RST304
        # no parameter_order, then nothing to check
        if target_cls.parameter_order is None:
            return None

        po_set = set(target_cls.parameter_order)
        pk = target_cls.parameters.keys()
        too_few = pk - po_set
        too_many = po_set - pk
        if too_few or too_many:
            return ParameterOrderError(
                target_cls=target_cls, too_few=too_few, too_many=too_many
            )


@ConstructorValue.register("base")
class BaseConstructorValue(ConstructorValue):
    """Vanilla way of type checking and instantiating :class:`~adorn.data.constructor.Constructor` instances"""  # noqa: B950

    @classmethod
    def _from_obj(
        cls, target_cls: Constructor, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Dict[Any, Any]:
        """Instantiate the arguments and an instance of :attr:`~adorn.data.constructor.Constructor.subclass`

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Dict[Any, Any]): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Any: An instance of :attr:`~adorn.data.constructor.Constructor.subclass`

        Raises:
            ParameterOrderError: specifies the additional and/or missing parameters
                in :attr:`~adorn.unit.complex.Complex.parameter_order`
        """  # noqa: B950, DAR401, DAR402, RST304
        type_str = obj.pop("type", None)
        paramterer_order_check = cls.check_parameter_order(target_cls=target_cls)
        if paramterer_order_check is not None:
            raise paramterer_order_check
        kwargs = dict()
        for k in target_cls.parameter_order or obj.keys():
            if k in obj:
                parameter = Parameter(
                    target_cls.parameters[k].annotation, target_cls, kwargs, k
                )
                kwargs[k] = orchestrator.from_obj(parameter, obj[k])
        obj["type"] = type_str
        return target_cls.constructor(**kwargs)

    @classmethod
    def _type_check(
        cls, target_cls: Constructor, orchestrator: "Orchestrator", obj: Dict[Any, Any]
    ) -> Optional[TypeCheckError]:
        """Check the arguments for :attr:`~adorn.data.constructor.Constructor.subclass`

        Exceptions:
            - :class:`~adorn.exception.type_check_error.ParameterOrderError`:
                specifies the additional and/or missing parameters
                in :attr:`~adorn.unit.complex.Complex.parameter_order`
            - :class:`~adorn.exception.type_check_error.KeyValueDiffError`:
                missing and/or additional arguments in ``obj``
            - :class:`~adorn.exception.type_check_error.KeyValueError`:
                args in ``obj`` didn't type check

        Args:
            target_cls (Constructor): instance of
                :class:`~adorn.data.constructor.Constructor`
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Dict[Any, Any]): an instance, containing the arguments for
                :attr:`~adorn.data.constructor.Constructor.subclass`

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` there are
                issues which would prevent creating the arguments
                for :attr:`~adorn.data.constructor.Constructor.subclass`,
                otherwise ``None``
        """  # noqa: RST304
        paramterer_order_check = cls.check_parameter_order(target_cls=target_cls)
        if paramterer_order_check is not None:
            return paramterer_order_check
        consistency_message = cls.parameter_consistency(target_cls, obj)
        if consistency_message:
            return consistency_message
        return cls.dict_type_check(
            target_cls=target_cls, orchestrator=orchestrator, obj=obj
        )



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Container of types with a class hierarchy."""
import inspect
import logging
from collections import defaultdict
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    List,
    Optional,
    Tuple,
    Type,
    TypeVar,
    Union,
    _GenericAlias,
    cast,
)

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.data.constructor import Constructor
from adorn.exception.configuration_error import ConfigurationError
from adorn.exception.type_check_error import (
    ComplexTypeMismatchError,
    ParamError,
    TypeCheckError,
    UnRepresentedTypeError,
)
from adorn.params import Params
from adorn.unit.unit import Unit

logger = logging.getLogger(__name__)

_T = TypeVar("_T")
_UnitT = TypeVar("_UnitT", bound="Unit")
_ComplexT = TypeVar("_ComplexT", bound="Complex")
_SubclassRegistry = Dict[str, Tuple[type, Optional[str]]]


class Complex(Unit):
    """:class:`~adorn.unit.unit.Unit` that preserves class hierarchy

    It is recommended you use this construct for your custom package.
    """

    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]]
    """maps a parent class to all of its subclasses and their associated names"""
    _subclass_registry: Dict[type, Tuple[type, str]]
    """classes that can be instantiated, mapped to their parent and name"""
    _intermediate_registry: DefaultDict[type, List[type]]
    """mapping of classes, who are parents, to their parents"""
    _parent_registry: DefaultDict[_ComplexT, List[_ComplexT]]
    """mapping of parents to their children, who are also parents"""

    constructor_type: Optional[str] = None
    """The name of the :class:`~adorn.unit.constructor_value.ConstructorValue`
       you want to instantiate the arguments for your class"""
    parameter_order: Optional[List[str]] = None
    """The order in which the
       parameters for a class should be instantiated"""
    constructor_name: Optional[str] = None
    """Name of the constructor to use to instantiate the class"""

    @classmethod
    def root(cls: Type[_ComplexT]) -> Callable[[Type[_T]], Type[_T]]:
        """Initialize all the class attributes needed to track the class hieracrchy

        This decorator is typically called on a direct descendant of
        :class:`~adorn.unit.complex.Complex` enabling all of the direct
        descendant's children to be tracked.

        Returns:
            Callable[[Type[_T]], Type[_T]]: decorator that initializes the
                class attributes for tracking
        """

        def _root(subclass: Type[_T]) -> Type[_T]:
            subclass._registry = defaultdict(dict)
            subclass._subclass_registry = dict()
            subclass._intermediate_registry = defaultdict(list)
            subclass._parent_registry = defaultdict(list)
            return subclass

        return _root

    @classmethod
    def register(
        cls: Type[_ComplexT], name: Optional[str], exist_ok: bool = False
    ) -> Callable[[Type[_T]], Type[_T]]:
        """Decorator that adds a class to the class hierarchy

        Args:
            name (Optional[str]): The ``str`` that will map to the given type, if
                ``None``, then the class is both a child of another ``Complex`` type
                and a parent of other ``Complex`` type(s), but cannot be
                instantiated
            exist_ok (bool): if ``True``, then `name` is allowed to already be in the
                `_registry`

        Returns:
            Callable[[Type[_T]], Type[_T]]: funcion that adds the decorated type
                to the ``_registry`` with the ``name`` being the key and the decorated
                class being the value
        """
        registry = cls._registry[cls]

        def add_subclass_to_registry(subclass: Type[_T]) -> Type[_T]:
            # Add to registry, raise an error if key has already been used.
            if name is not None:
                if not exist_ok and name in registry:
                    message = (
                        f"Cannot register {name} as {cls.__name__}; "
                        f"name already in use for {registry[name].__name__}"
                    )
                    raise ConfigurationError(message)
                registry[name] = subclass
                cls._subclass_registry[subclass] = (cls, name)
                if cls in cls._subclass_registry:
                    parent, _ = cls._subclass_registry[cls]
                    cls._parent_registry[parent].append(cls)
            else:
                cls._intermediate_registry[subclass].append(cls)
                cls._parent_registry[cls].append(subclass)
            return subclass

        return add_subclass_to_registry

    @classmethod
    def get_instantiate_children(cls: Type[_ComplexT]) -> Dict[str, _ComplexT]:
        """Get the subclasses for a class that can be instantiated

        Returns:
            Dict[str, _ComplexT]: All the subclasses of a class that can be instantiated
        """
        instantiate_children = cls._explode_registry()
        for i in cls.get_intermediate_children():
            instantiate_children = {
                **i.get_instantiate_children(),
                **instantiate_children,
                **i._explode_registry(),
            }
        return instantiate_children

    @classmethod
    def _explode_registry(cls: Type[_ComplexT]) -> Dict[str, _ComplexT]:
        """Helper method, to get all the descendants of a class

        Returns:
            Dict[str, _ComplexT]: Dictionary containing all the descendants of a class
        """
        instantiate_children = (
            dict() if cls not in cls._subclass_registry else {cls.name(): cls}
        )
        for v in cls._registry[cls].values():
            instantiate_children = {
                **instantiate_children,
                **v.get_instantiate_children(),
            }
        return instantiate_children

    @classmethod
    def get_direct_parent(
        cls: Type[_ComplexT],
    ) -> Optional[Union[_ComplexT, List[_ComplexT]]]:
        """Get the parent(s) of a class

        Returns:
            Optional[Union[_ComplexT, List[_ComplexT]]]: if ``None``, then the class
                has no parents, if ``_ComplexT`` then the class has a parent,
                if ``List[_ComplexT]`` then the class has multiple parents
        """
        if cls in cls._parent_registry:
            return None

        if cls in cls._intermediate_registry:
            return cls._intermediate_registry[cls]

        return cls._subclass_registry[cls][0]

    @classmethod
    def get_intermediate_children(cls: Type[_ComplexT]) -> List[_ComplexT]:
        """Get the children of a class, that are also parents

        .. note::
            Some of the classes in the list may not be able to be instantiated

        Returns:
            List[_ComplexT]: list of subclasses of that are also parents
        """
        return cls._parent_registry.get(cls, [])

    @classmethod
    def name(cls: Type[_ComplexT]) -> str:
        """Get the string associated with a given class

        Returns:
            str: the string associated with a class

        Raises:
            Exception: the given class, cannot be instantiated and therefore
                does not have a name
        """
        if cls not in cls._subclass_registry:
            raise Exception(f"{cls} was not registered as a constructable subclass")
        return cls._subclass_registry[cls][1]

    @classmethod
    def by_name(
        cls: Type[_ComplexT], name: Optional[str] = None
    ) -> Callable[..., _ComplexT]:
        """Returns a callable function that constructs an argument of the class.

        Args:
            name (Optional[str]): name of the class you want the constructor for,
                if ``None``, :meth:`~adorn.unit.complex.Complex.name` will be used

        Returns:
            Callable[..., _ComplexT]: constructor for the requested class
        """
        if cls not in cls._subclass_registry:
            return cls
        name = name or cls.name()
        logger.info(f"instantiating registered subclass {name} of {cls}")
        subclass = cls.resolve_class_name(name)
        return cast(Type[_ComplexT], subclass)

    @classmethod
    def resolve_class_name(
        cls: Type[_ComplexT], name: Optional[str] = None
    ) -> Type[_ComplexT]:
        """Returns the class that corresponds to the given `name`

        Args:
            name (str): name of the class the caller wants

        Returns:
            Type[_ComplexT]: the type associated with ``name`` specified
        """
        name = name or cls.name()
        children = cls.get_instantiate_children()
        return children[name]

    @classmethod
    def list_available(cls: Type[_ComplexT]) -> List[str]:
        """List all the names of subclasses that can be instantiated

        Returns:
            List[str]: names of subclasses that can be instantiated
        """
        return list(cls.get_instantiate_children().values())

    def match(self, key: Any) -> bool:
        """Check if the key is a subclass of ``Complex``

        Args:
            key (Any): potential type to be checked

        Returns:
            bool: if ``True`` then key is a Type of the appropriate type
        """
        return (
            (not isinstance(key, _GenericAlias))
            and inspect.isclass(key)
            and issubclass(key, type(self))
        )

    def get(self, key: Any, orchestrator: "Orchestrator") -> Optional[type]:
        """Return the relevant registered type

        Args:
            key (Any): the type that may be a registered value
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[type]: if ``key`` is a registered value,
                then the registered value is returned, otherwise ``None``
        """
        if self.match(key):
            return key._get()
        return None

    @classmethod
    def _get(cls: Type[_ComplexT]) -> Optional[type]:
        """Get the relevant class

        This method exists to faciliate potential future use cases
        that may require additional logic for getting the correct class

        Returns:
            Optional[type]: return the class associated with the classmethod
        """
        return cls

    def contains(self, obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` is registered

        Args:
            obj (Any): potentially a type that may be registered
            orchestrator (Orchestrator): container of all types, which
                is used for nested types

        Returns:
            bool: if ``True``, the type is registered
        """
        if self.match(obj):
            return obj._contains(obj, orchestrator)
        return False

    @classmethod
    def _contains(cls: Type[_ComplexT], obj: Any, orchestrator: "Orchestrator") -> bool:
        """Check if the ``cls`` has been registered

        Args:
            obj (Any): this is unused
            orchestrator (Orchestrator): this is unused

        Returns:
            bool: if ``True``, the type can be represented by ``cls``
        """
        return cls in cls._subclass_registry or cls in cls._registry

    def general_check(
        self,
        cls: Type,
        orchestrator: "Orchestrator",
        obj: Any,
        is_from_obj: bool = False,
    ) -> Optional[TypeCheckError]:
        """Checks that need to be performed for both ``type_check`` and ``from_obj``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.ParamError`:
                ``obj`` was not of type :class:`~adorn.params.Params`
            - :class:`~adorn.exception.type_check_error.ComplexTypeMismatchError`:
                The ``type`` key contained in ``obj`` had a value that did not
                map to any of the valid subclasses of ``target_cls``

        Args:
            cls (Type): the target class that ``obj`` is trying to be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): the instance that will have the sanity checks applied
                to it
            is_from_obj (bool): if ``True`` this check is being done for
                :meth:`~adorn.unit.complex.Complex.from_obj`, which allows
                instances of ``cls`` to be considered valid

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(cls, orchestrator):
            return UnRepresentedTypeError(cls, self, obj)
        if is_from_obj and isinstance(obj, cls):
            return obj
        if not isinstance(obj, Params):
            return ParamError(target_cls=cls, obj=obj)

        children = cls.get_instantiate_children()
        if obj.get("type") not in children:
            return ComplexTypeMismatchError(cls, children, obj)

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.ParamError`:
                ``obj`` was not of type :class:`~adorn.params.Params`
            - :class:`~adorn.exception.type_check_error.ComplexTypeMismatchError`:
                The ``type`` key contained in ``obj`` had a value that did not
                map to any of the valid subclasses of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be checked to see if it can be converted to an
                instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        general_check_output = self.general_check(target_cls, orchestrator, obj)
        if general_check_output is not None:
            return general_check_output
        ru = target_cls.resolve_class_name(obj.get("type"))
        return ru._type_check(target_cls, orchestrator, obj)

    @classmethod
    def _type_check(
        cls: Type[_ComplexT], target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Check that the args, specified in ``obj``, for ``target_cls`` type check

        The checking of arguments for a ``target_cls`` is delegated to
        :class:`~adorn.unit.constructor_value.ConstructorValue`

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be checked to see if it can be converted to an
                instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        return orchestrator.type_check(Constructor(cls), obj)

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Convert `obj` to an instance of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be converted into an instance of ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` derived from ``obj``

        Raises:
            UnRepresentedTypeError: ``obj`` did not map to any of the types contained in
                the ``_registry``
            ParamError: ``obj`` was not of type :class:`~adorn.params.Params`
            ComplexTypeMismatchError: The ``type`` key contained in ``obj`` had a value
                that did not map to any of the valid subclasses of ``target_cls``
        """  # noqa: DAR401, DAR402
        general_check_output = self.general_check(
            target_cls, orchestrator, obj, is_from_obj=True
        )
        if isinstance(general_check_output, TypeCheckError):
            raise general_check_output
        if general_check_output is not None:
            return general_check_output

        ru = target_cls.resolve_class_name(obj.get("type"))
        return ru._from_obj(target_cls, orchestrator, obj)

    @classmethod
    def _from_obj(
        cls: Type[_ComplexT], target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Any:
        """Convert the args, specified in ``obj``, to instances to be passed ``target_cls`` constructor

        The instantiation of arguments for the``target_cls`` and instantiation of the ``target_cls``
        are delegated to :class:`~adorn.unit.constructor_value.ConstructorValue`

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, usually of type :class:`~adorn.params.Params`,
                that will be converted into an instance of ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` derived from ``obj``
        """  # noqa: B950
        if isinstance(obj, cls):
            return obj
        return orchestrator.from_obj(Constructor(cls), obj)



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Exceptions specifying why a request was malformed"""
from typing import TYPE_CHECKING, Any, Dict, List, Optional, Set, Tuple, Type, Union

if TYPE_CHECKING:  # pragma: no cover
    from adorn.data.parameter import Parameter
    from adorn.params import Params
    from adorn.unit.anum import Anum
    from adorn.unit.unit import Unit


DictOrList = Union[
    Dict[Union[int, str], "TypeCheckError"],
    List[Tuple[Dict[Union[int, str], "TypeCheckError"]]],
]


class TypeCheckError(Exception):
    """Generic ``Exception`` that contains the state the produced the error.

    Args:
        target_cls (Type): the requested type to perform an action against
        msg (List[str]): explanation about the exception
        child (Optional[TypeCheckError]): an excpetion produced by a downstream
            process
        obj (Optional[Any]): the caller provided object, that was innvolved in
            causing the error
    """

    def __init__(
        self,
        target_cls: Type,
        msg: List[str],
        child: Optional["TypeCheckError"] = None,
        obj: Optional[Any] = None,
    ) -> None:
        super().__init__()
        self.target_cls = target_cls
        self.msg = msg
        self.child = child
        self.obj = obj

    def to_str(self, seed: str = "") -> str:
        """Create a sting with the reason/explanation of the error.

        Args:
            seed (str): spacing used to compose error statements in a cute way

        Returns:
            str: information about the error
        """
        prefix = "\n" + seed if seed else ""
        line_break = "\n" + seed
        outer_msg = prefix + line_break.join(self.msg)
        if self.child is None:
            return outer_msg
        return outer_msg + self.child.to_str(seed + "\t")

    def __str__(self) -> str:
        return self.__repr__()

    def __repr__(self) -> str:
        return self.to_str()

    def __eq__(self, other) -> bool:
        if type(self) != type(other):  # noqa: E721
            return False
        return all(
            [
                self.target_cls == other.target_cls,
                self.msg == other.msg,
                self.child == other.child,
                self.obj == other.obj,
            ]
        )


class WrongTypeError(TypeCheckError):
    """Error when provided an ``obj`` of the wrong type

    Args:
        target_cls (type): the requested type to perform an action against
        obj (Any): caller provided object, that is not of type
            ``target_cls``
    """

    def __init__(self, target_cls: type, obj: Any) -> "TypeCheckError":
        target_cls_name = (
            target_cls.__name__ if hasattr(target_cls, "__name__") else target_cls
        )
        msg = [
            f"Expected an object of type {target_cls_name},\n ",
            "but received an object of type ",
            f"{type(obj).__name__}\nwith a value of:",
            f"\n\t{obj}",
        ]
        super().__init__(target_cls=target_cls, msg=msg, child=None, obj=obj)


class ParamError(TypeCheckError):
    """Error when an obj of not type :class:`~adorn.params.Params` is passed

    Args:
        target_cls (type): the requested type to perform an action against
        obj (Any): caller provided object, that is not of type
            ``Params``
    """

    def __init__(self, target_cls: type, obj: Any) -> "TypeCheckError":
        target_cls_name = (
            target_cls.__name__ if hasattr(target_cls, "__name__") else target_cls
        )
        msg = [
            f"For {target_cls_name}, expected a Params object",
            f"but received an object of type {type(obj).__name__}",
        ]
        super().__init__(target_cls=target_cls, msg=msg, child=None, obj=obj)


class UnRepresentedTypeError(TypeCheckError):
    """Error when a :class:`~adorn.unit.unit.Unit` does not cotain a given type.

    Args:
        target_cls (type): The requested type that can't be expressed by the ``Unit``
        unit (Unit): a collection of types that can't express ``target_cls``
        obj (Any): the caller created object that was supposed to map to ``target_cls``
    """

    def __init__(self, target_cls: type, unit: "Unit", obj: Any) -> "TypeCheckError":
        target_cls_name = (
            target_cls.__name__ if hasattr(target_cls, "__name__") else target_cls
        )
        msg = [
            (
                f"Requested type: {target_cls_name} with an object of type:"
                f"{type(obj).__name__}"
            ),
            "with a value of:",
            f"{obj}",
            f"didn't match any subtype of {type(unit).__name__}",
        ]
        super().__init__(target_cls=target_cls, msg=msg, child=None, obj=obj)


class KeyValueError(TypeCheckError):
    """Error when dependencies of the given ``target_cls`` produce an error.

    Args:
        target_cls (type): the requested type, whose dependencies produced error(s)
        key_values (DictOrList): a collection of the name of the dependency and
            error the dependency produced
        obj (Union[List[Any], Dict[str, Any]]): caller specified values for the
            dependencies
    """

    def __init__(
        self,
        target_cls: type,
        key_values: DictOrList,
        obj: Union[List[Any], Dict[str, Any]],
    ) -> None:
        msg = self.get_msg(target_cls, key_values)
        self.key_values = key_values
        super().__init__(target_cls, msg, child=None, obj=obj)

    @staticmethod
    def get_msg(
        target_cls: type,
        key_values: DictOrList,
    ) -> List[str]:
        """Generate the msg containing why the dependencies produced an error.

        Args:
            target_cls (type): the requested type, whose dependencies produced
                error(s)
            key_values (DictOrList): a collection of the name of the dependency
                and error the dependency produced

        Returns:
            List[str]: message containing the reasons why the dependencies produced
                errors
        """
        result_list = []
        collection = key_values.items() if isinstance(key_values, dict) else key_values
        for k, v in collection:
            msg = "\t-" + f"{k}: " + v.to_str("\t\t")
            result_list.append(msg)
        return [
            f"Failed to construct {target_cls} because of the following values:"
        ] + result_list


class HashableError(TypeCheckError):
    """Error when a Unit's subtype is not hashable.

    This is typically caused when a type is wrapped in a ``Set``, but the
    wrapped type is not hashable.

    Args:
        target_cls (Type): The requested type, that isn't hashable
        unit (Unit): the collection of types that contained ``target_cls``
        obj (Optional[Any]): the caller created object that maps
            to the unhashable type
    """

    def __init__(self, target_cls: Type, unit: "Unit", obj: Optional[Any]) -> None:
        target_origin = getattr(target_cls, "__origin__", None)
        target_arg = getattr(target_cls, "__args__", [None])[0]
        ta = target_arg or target_origin or target_cls
        msg = [
            f"Failed to construct {target_cls}, because {ta}",
            "is not hashable.  You may need to alter the hashable attr of",
            f"{unit.__name__} or use a type that is hashable",
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)


class TupleArgLenError(TypeCheckError):
    """Exception when the number of args specified in the Tuple does not match the object's length.

    Args:
        target_cls (Type): the tuple type
        obj (Optional[Any]): the caller created object that has a length
            different than that specified by ``target_cls``
    """  # noqa: B950

    def __init__(self, target_cls: Type, obj: Optional[Any]) -> None:
        target_args = getattr(target_cls, "__args__", [None])
        msg = [
            f"Failed to create a {target_cls} because {len(target_args)} args were",
            f"expected but only {len(obj)} were received.",
            f"obj had a value of {obj}",
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)


class KeyValueDiffError(TypeCheckError):
    """:class:`~adorn.params.Params` object had missing and/or additional information

    Args:
        target_cls (type): the type that the ``obj`` was intended to create
        parameter_kv (Optional[Dict[str, type]]): missing information from the ``obj``
        obj_kv (Optional[Dict[str, type]]): additional information in the ``obj``
        obj (Params): the ``Params`` object that has missing and/or additional
            information
    """

    def __init__(
        self,
        target_cls: type,
        parameter_kv: Optional[Dict[str, type]],
        obj_kv: Optional[Dict[str, type]],
        obj: "Params",
    ) -> None:
        msg = [
            f"{target_cls} was missing arguments, indicated with a `-`",
            "and/or passed additional arguments indicated with a `+`",
        ]
        if parameter_kv:
            missing = [
                f"\t- {k}: {getattr(v, '__name__', v)}" for k, v in parameter_kv.items()
            ]
            msg.extend(missing)

        if obj_kv:
            added = [f"\t+ {k}: {getattr(v, '__name__', v)}" for k, v in obj_kv.items()]
            msg.extend(added)
        self.parameter_kv = parameter_kv
        self.obj_kv = obj_kv
        super().__init__(target_cls, msg, child=None, obj=obj)


class ComplexTypeMismatchError(TypeCheckError):
    """The `type` specified doesn't match any of the options for the ``target_cls``

    Args:
        target_cls (Type): the requested type
        possibilities (Dict[str, type]): the different subclasses that can be
            requested for ``target_cls``
        obj (Params): ``Params`` with a value for ``type`` that doesn't map to
            any of the subclasses of ``target_cls``
    """

    def __init__(
        self, target_cls: Type, possibilities: Dict[str, type], obj: "Params"
    ) -> None:
        msg = [
            f"For {target_cls}, a Params object specified a `type`",
            f"of {obj.get('type')}, which is not an acceptable option",
            f"for {target_cls}.  The potential values for `type` and",
            "their corresponding type are:",
            *[f"\t- {k}: {v.__name__}" for k, v in possibilities.items()],
        ]
        self.possiblilites = possibilities
        super().__init__(target_cls, msg, child=None, obj=obj)


class ParameterOrderError(TypeCheckError):
    """The ``parameter_order`` attr contained too many and/or too few parameters

    Args:
        target_cls (type): the type with a malformed ``parameter_order`` attr
        too_few (Set[str]): parameters not in ``parameter_order``
        too_many (Set[str]): additional parameters in ``parameter_order`` that aren't
            in the constructor
    """

    def __init__(
        self,
        target_cls: type,
        too_few: Set[str],
        too_many: Set[str],
    ) -> None:
        msg = [
            f"{target_cls} parameter_order attribute, was missing parameters,",
            "indicated with a `-` and/or passed additional arguments",
            "indicated with a `+`",
        ]
        if too_few:
            missing = [f"\t- {k}" for k in too_few]
            msg.extend(missing)

        if too_many:
            added = [f"\t+ {k}" for k in too_many]
            msg.extend(added)

        self.too_few = too_few
        self.too_many = too_many
        super().__init__(target_cls, msg, child=None, obj=None)


class MalformedDependencyError(TypeCheckError):
    """:class:`~adorn.unit.parameter_value.Dependent` type was not provided all the necessary information.

    Args:
        target_cls (Type): The type with malformed dependency information
    """  # noqa: B950

    def __init__(self, target_cls: Type) -> None:
        msg = [
            f"{target_cls.cls} type was not provided all the necessary",
            "information at the type level.  Ensure that all necessary",
            "information was passed to the type.",
        ]
        super().__init__(target_cls, msg, child=None, obj=None)


class MissingLiteralError(TypeCheckError):
    """:class:`~adorn.unit.parameter_value.Dependent` type's first argument was not a ``typing.Literal``.

    Args:
        target_cls (Type): Type with a dependency that didn't have a ``typing.Literal``
    """  # noqa: B950

    def __init__(self, target_cls: Type) -> None:
        msg = [
            f"{target_cls.cls} requires it's last argument to be wrapped",
            "in a ``typing.Literal``",
        ]
        super().__init__(target_cls, msg, child=None, obj=None)


class UnaryLiteralError(TypeCheckError):
    """``typing.Literal`` that was passed more than one arg

    Args:
        target_cls (Type): Type with a dependency that had a ``typing.Literal``
            with more than one arg
    """

    def __init__(self, target_cls: Type) -> None:
        msg = [
            f"{target_cls.cls} requires it's last argument to be wrapped",
            "in a ``typing.Literal`` which is passed a single argument.",
        ]
        super().__init__(target_cls, msg, child=None, obj=None)


class MalformedLiteralError(TypeCheckError):
    """``typing.Literal`` argument wasn't of the correct type

    Args:
        target_cls (Type): the type which contained the problem ``typing.Literal``
        literal_type (Type): the required type of the ``typing.Literal``
            argument
        child (TypeCheckError): the error produced by the ``typing.Literal``
            argument
    """

    def __init__(
        self, target_cls: Type, literal_type: Type, child: TypeCheckError
    ) -> None:
        msg = [
            f"{target_cls.cls} requires it's last argument to be wrapped",
            f"in a ``typing.Literal`` which is of type {literal_type}",
        ]
        self.literal_type = literal_type
        super().__init__(target_cls, msg, child=child, obj=None)


class ExtraLiteralError(TypeCheckError):
    """``typing.Literal`` contains additional keys not specified in the constructor.

    Args:
        target_cls (Type): the type which contains the problem ``typing.Literal``
        extras (List[str]): the additional arguments specified in the
            ``typing.Literal`` argument
    """

    def __init__(self, target_cls: Type, extras: List[str]) -> None:
        msg = [
            f"{target_cls.cls}'s' ``typing.Literal`` arg contains additional keys",
            "that were not in the constructor",
            *[f"\t- {k}" for k in extras],
        ]
        self.extras = extras
        super().__init__(target_cls, msg, child=None, obj=None)


class TooDeepLiteralError(TypeCheckError):
    """``typing.Literal`` specified a dependency that goes beyond one layer deep

    An example would be `"a.b.c"`.

    Args:
        target_cls (Type): the type which contains the problem ``typing.Literal``
        bad_literal (List[str]): the dependency requests that went too deep
    """

    def __init__(self, target_cls: Type, bad_literal: List[str]) -> None:
        msg = [
            f"{target_cls.cls}'s' ``typing.Literal`` arg is not allowed to have",
            "dependencies that go beyond one layer deep, one period.",
            "The following dependencies were more than one layer deep:",
            *[f"\t- {k}" for k in bad_literal],
        ]
        self.bad_literal = bad_literal
        super().__init__(target_cls, msg, child=None, obj=None)


class MissingDependencyError(TypeCheckError):
    """``typing.Literal`` requested a dependency that doesn't exist in local state

    Args:
        target_cls (Type): the type which contains the problem ``typing.Literal``
        missing_dependency (Dict[str, str]): the dependency requests that didn't
            exist in the local state
    """

    def __init__(self, target_cls: Type, missing_dependency: Dict[str, str]) -> None:
        msg = [
            f"{target_cls.cls}'s' ``typing.Literal`` requested",
            "dependencies that were not in the local state.",
            "The following dependency requests were not in the local state:",
            *[f"\t- {k}: {v}" for k, v in missing_dependency.items()],
        ]
        self.missing_dependency = missing_dependency
        super().__init__(target_cls, msg, child=None, obj=None)


class AnumWrongTypeError(WrongTypeError):
    """Error when provided an ``obj`` of the wrong type for a :class:`~adorn.unit.anum.Anum`

    Args:
        target_cls (Anum): the requested wrapped enumeration
        obj (Any): caller provided object, that is not of type
            ``target_cls``
    """  # noqa: B950

    def __init__(self, target_cls: "Anum", obj: Any) -> "TypeCheckError":
        super().__init__(str, obj)
        self.target_cls = target_cls
        self.msg[0] = f"For the Anum, {self.target_cls}, " + self.msg[0]


class AnumMemberError(TypeCheckError):
    """The str specified doesn't match any of the members of ``target_cls``

    Args:
        target_cls (Anum): the requested wrapped enumeration
        obj (str): requested member that does not exist
    """

    def __init__(self, target_cls: "Anum", obj: str) -> None:
        self.members = target_cls.__members__
        msg = [
            f"For {target_cls}, a str specified a member,",
            f"{obj}, which is not an acceptable option",
            f"for {target_cls}.  The valid members",
            f"of {target_cls} are:",
            *[f"\t- {k}" for k in self.members.keys()],
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)


class UserDictError(TypeCheckError):
    """Report an error during the from_obj step of :class:`~adorn.alter.dict_alter.UserDictAlter`

    Args:
        alter_type (Type): the type of ``Alter`` being performed
        target_cls (Parameter): the parameter the alter was trying to convert
            the ``obj`` into
        obj (Any): the alter request that produced an exception
        exception (Exception): the ``Exception`` produced from trying to convert ``obj``
            into the requested data.

    """  # noqa: B950

    def __init__(
        self, alter_type: Type, target_cls: "Parameter", obj: Any, exception: Exception
    ) -> None:
        self.alter_type = alter_type
        self.exception = exception
        msg = [
            f"An Alter of type {alter_type} was requested for",
            f"a parameter named {target_cls.parameter_name} of type {target_cls.cls}",
            "but an exception was caused when converting the obj:",
            f"{obj}",
            f"to type {target_cls.cls}.  The exception was:",
            f"{self.exception}",
        ]
        super().__init__(target_cls, msg, child=None, obj=obj)

    def __eq__(self, other) -> bool:
        if type(self) != type(other):  # noqa: E721
            return False
        return all(
            [
                self.target_cls == other.target_cls,
                self.msg[:-1] == other.msg[:-1],
                self.child == other.child,
                self.obj == other.obj,
                isinstance(self.exception, type(other.exception)),
            ]
        )



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""A container of information relevant for constructing an object."""
import inspect
from dataclasses import dataclass
from typing import TYPE_CHECKING, Callable, Dict, Type

if TYPE_CHECKING:  # pragma: no cover
    from adorn.unit.complex import _T, _UnitT


@dataclass
class Constructor:
    """Container holding the class and relevant metadata to construct an object.

    Args:
        subclass (_UnitT): The ``type`` to generate construction metadata for
    """

    subclass: "_UnitT"

    def __post_init__(self):
        self.constructor_type = getattr(self.subclass, "constructor_type", None)
        if hasattr(self.subclass, "by_name"):
            self.constructor = self.subclass.by_name()
            constructor = (
                self.constructor
                if self.constructor.constructor_name is None
                else getattr(self.constructor, self.constructor.constructor_name)
            )
        else:
            self.constructor = self.subclass
            constructor = self.constructor
        self.parameters = self.infer_params(self.subclass, constructor)
        self.parameter_order = getattr(self.subclass, "parameter_order", None)

    def infer_params(
        self, target_class: Type["_T"], constructor: Callable[..., "_T"]
    ) -> Dict[str, inspect.Parameter]:
        """Generate a dict containing all the parameters needed to instantiate ``subclass``.

        This method is recursive and will convert ``**kwargs``, specified in the constructor,
        to the relevant parameters of its parent class(es), given the parent(s)
        subclass :class:`~adorn.unit.complex.Complex`

        Args:
            target_class (Type["_T"]): The class you are interegating the constructor of.
                This argument is used to explode the ``**kwargs``
            constructor (Callable[..., "_T"]): The method, whose parameters will be added
                to a dictionary

        Returns:
            Dict[str, inspect.Parameter]: constructor parameters as a dict
                keys are the parameter names and values are the parameter metadata
        """  # noqa:B950
        # circular imports
        from adorn.unit.complex import Complex

        signature = inspect.signature(constructor)
        parameters = dict(signature.parameters)

        parameters.pop("self", None)

        has_kwargs = False
        var_positional_key = None
        for param in parameters.values():
            if param.kind == param.VAR_KEYWORD:
                has_kwargs = True
            elif param.kind == param.VAR_POSITIONAL:
                var_positional_key = param.name

        if var_positional_key:
            del parameters[var_positional_key]

        if not has_kwargs:
            return parameters
        # probably only want to do this when the associated type is inspect._empty
        parameters.pop("kwargs")
        # "mro" is "method resolution order".  The first one is the current class,
        # the next is the first superclass, and so on.  We take the first
        # superclass we find that inherits from Complex.
        super_class = None
        # we could probably use the Complex's parent method
        # to help filter the mro output
        for super_class_candidate in target_class.mro()[1:]:
            if issubclass(super_class_candidate, Complex):
                super_class = super_class_candidate
                break
        if super_class:
            super_parameters = self.infer_params(super_class, super_class.by_name())
        else:
            super_parameters = {}

        return {
            **super_parameters,
            **parameters,
        }  # Subclass parameters overwrite superclass ones



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Representation of ``enum.Enum``"""
from enum import Enum
from inspect import isclass
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    Optional,
    Type,
    TypeVar,
)

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.type_check_error import (
    AnumMemberError,
    AnumWrongTypeError,
    TypeCheckError,
    UnRepresentedTypeError,
)
from adorn.unit.unit import Unit

_E = TypeVar("_E", bound=Enum)


class Anum(Unit):
    """A wrapper around :class:`~enum.Enum`

    The members of the enumeration are added to
    :attr:`~adorn.unit.anum.Anum._registry` and are accessible
    by specifying the enumeration they belong to and the ``str``
    representation of their name

    .. code-block:: python

        from collections import defaultdict
        from enum import auto, Enum

        from adorn.orchestrator.base import Base
        from adorn.unit.anum import Anum

        class Example(Anum):
            _registry = defaultdict(dict)

        @Example.register()
        class AB(Enum):
            a = auto
            b = auto

        base = Base([Example()])
        assert base.from_obj(AB, "a") == AB.a

    """  # noqa: RST304

    _registry: ClassVar[DefaultDict[type, Dict[str, type]]]

    @classmethod
    def register(cls) -> Callable[[Type[_E]], Type[_E]]:  # pragma: no cover
        """Add :class:`~enum.Enum` and it's members to the :attr:`~adorn.unit.anum.Anum._registry`

        Returns:
            Callable[[Type[_E]], Type[_E]]: funcion that adds the decorated
                :class:`~enum.Enum` and its members to the ``_registry``
        """  # noqa: B950, D202, RST304

        def add_members(subclass: Type[_E]) -> Type[_E]:
            registry = cls._registry[subclass]
            for k, v in subclass.__members__.items():
                registry[k] = v
            return subclass

        return add_members

    def get(
        self, key: Any, orchestrator: "Orchestrator"
    ) -> Optional[type]:  # pragma: no cover
        """Return the relevant ``Anum`` for the `key`

        Args:
            key (Any): the Type that may be a subclass of the ``Unit``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            Optional[type]: if ``key`` is an ``Anum`` then return it,
                otherwise ``None``
        """
        if self.contains(key, orchestrator):
            return key

    def contains(
        self, obj: Any, orchestrator: "Orchestrator"
    ) -> bool:  # pragma: no cover
        """Check if `obj` is registered to ``Anum``

        Args:
            obj (Any): the Type that may be a subclass of ``Anum``
            orchestrator (Orchestrator): container of all types, not used

        Returns:
            bool: if ``True`` then ``obj`` is represented by ``Anum``,
                otherwise ``False``
        """
        return isclass(obj) and issubclass(obj, Enum) and obj in self._registry

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:  # pragma: no cover
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``
            - :class:`~adorn.exception.type_check_error.AnumWrongTypeError`:
                ``obj`` was not of type ``str``
            - :class:`~adorn.exception.type_check_error.AnumMemberError`:
                ``obj`` specified a value that did not map to one of the
                members of ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, not used
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(target_cls, orchestrator):
            return UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)

        if not isinstance(obj, str):
            return AnumWrongTypeError(target_cls=target_cls, obj=obj)

        if obj not in self._registry[target_cls]:
            return AnumMemberError(target_cls=target_cls, obj=obj)

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> _E:
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into
            orchestrator (Orchestrator): container of all types, not used
            obj (Any): an instance, that will be converted to an
                instance of ``target_cls``

        Returns:
            _E: a member of ``target_cls``
        """
        return self._registry[target_cls][obj]



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Container of types with a flat class hierarchy."""
import logging
from typing import (
    TYPE_CHECKING,
    Any,
    Callable,
    ClassVar,
    DefaultDict,
    Dict,
    Iterable,
    Optional,
    Type,
    TypeVar,
)

if TYPE_CHECKING:  # pragma: no cover
    from adorn.orchestrator.orchestrator import Orchestrator

from adorn.exception.configuration_error import ConfigurationError
from adorn.exception.type_check_error import TypeCheckError, UnRepresentedTypeError
from adorn.unit.unit import Unit

logger = logging.getLogger(__name__)

_T = TypeVar("_T")
A = TypeVar("A")
_SubclassRegistry = Dict[str, type]


class Simple(Unit):
    """:class:`~adorn.unit.unit.Unit` for types with a flat class hierarchy

    Attributes:
        _registry (ClassVar[DefaultDict[type, _SubclassRegistry]]): contains mapping
            from ``str`` to ``Type``
    """

    _registry: ClassVar[DefaultDict[type, _SubclassRegistry]]

    @classmethod
    def register(cls, name: str) -> Callable[[Type[_T]], Type[_T]]:
        """Decorator that adds a type to the ``_registry``

        Args:
            name (str): The ``str`` that will map to the given type

        Returns:
            Callable[[Type[_T]], Type[_T]]: funcion that adds the decorated type
                to the ``_registry`` with the ``name`` being the key and the decorated
                class being the value
        """
        registry = cls._registry[cls]

        def add_subclass_to_registry(subclass: Type[_T]) -> Type[_T]:
            # Add to registry, raise an error if key has already been used.
            if name in registry:
                message = (
                    f"Cannot register {name} as {cls.__name__}; "
                    f"name already in use for {registry[name].__name__}"
                )
                raise ConfigurationError(message)
            registry[name] = subclass
            return subclass

        return add_subclass_to_registry

    def _get(self, key: type, orchestrator: "Orchestrator") -> Optional[type]:
        """Return the relevant type from the ``_registry``

        Args:
            key (type): the type that may be a value in the
                ``_registry``
            orchestrator (Orchestrator): container of all types, typically
                used to recurse nested types

        Returns:
            Optional[type]: if ``key`` is a value in the ``_registry``,
                then the value is returned, otherwise ``None``
        """
        try:
            return next(self.relevant_types(key, orchestrator))
        except StopIteration:
            return None

    @classmethod
    def _contains(cls, obj: Type, orchestrator: "Orchestrator") -> bool:
        """Check if the ``obj`` can be represented by ``cls``

        .. note::

            This method should typically be overriden by all subclasses of
            ``Simple``.

        Args:
            obj (Type): the type that may be represented by ``cls``
            orchestrator (Orchestrator): container of all types, which
                is used for nested types

        Returns:
            bool: if ``True``, the type can be represented by ``cls``
        """
        return False

    def type_check(
        self, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Ensure `obj` can be converted to the type ``target_cls``

        Exceptions:
            - :class:`~adorn.exception.type_check_error.UnRepresentedTypeError`:
                ``obj`` did not map to any of the types contained in
                the ``_registry``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        if not self.contains(target_cls, orchestrator):
            return UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)
        ru = self.get(target_cls, orchestrator)
        return ru._type_check(target_cls, orchestrator, obj)

    @classmethod
    def _type_check(
        cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any
    ) -> Optional[TypeCheckError]:
        """Ensure `obj` can be converted to the type ``target_cls``

        .. note::

            This method should typically be overriden by all subclasses of
            ``Simple``.

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be checked
                against
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be checked to see if it
                can be converted to an instance of ``target_cls``

        Returns:
            Optional[TypeCheckError]: if ``TypeCheckError`` then there was
                an issue which would prevent converting ``obj`` to an instance
                of ``target_cls``, otherwise ``None``
        """
        return None

    def from_obj(self, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Convert `obj` to an instance of the type ``target_cls``

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into an instance of
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be converted to an instance of
                ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` that was derived from ``obj``

        Raises:
            UnRepresentedTypeError: ``obj`` did not map to any of the types contained in
                the ``_registry``
        """
        if not self.contains(target_cls, orchestrator):
            raise UnRepresentedTypeError(target_cls=target_cls, unit=self, obj=obj)
        ru = self.get(target_cls, orchestrator)
        return ru._from_obj(target_cls, orchestrator, obj)

    @classmethod
    def _from_obj(cls, target_cls: Type, orchestrator: "Orchestrator", obj: Any) -> Any:
        """Convert `obj` to an instance of the type ``target_cls``

        .. note::

            This method should typically be overriden by all subclasses of
            ``Simple``.

        Args:
            target_cls (Type): the target Type, the given ``obj`` will be converted
                into an instance of
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types
            obj (Any): an instance, that will be converted to an instance of
                ``target_cls``

        Returns:
            Any: an instance of ``target_cls`` that was derived from ``obj``
        """
        return None

    def relevant_types(self, key: Any, orchestrator: "Orchestrator") -> Iterable[Type]:
        """Iterable containing the types that map to the given ``key``

        Args:
            key (Any): the object to check if it maps to any types in the
                ``_registry``
            orchestrator (Orchestrator): container of all types, typically used
                to recurse down nested types

        Returns:
            Iterable[Type]: collection of types in the ``registry`` that map to ``key``
        """
        return (
            v_prime
            for v in self._registry.values()
            for v_prime in v.values()
            if v_prime._contains(key, orchestrator)
        )



# Copyright 2021 Jacob Baumbach
#
# Licensed under the Apache License, Version 2.0 (the "License");
# you may not use this file except in compliance with the License.
# You may obtain a copy of the License at
#
#     http://www.apache.org/licenses/LICENSE-2.0
#
# Unless required by applicable law or agreed to in writing, software
# distributed under the License is distributed on an "AS IS" BASIS,
# WITHOUT WARRANTIES OR CONDITIONS OF ANY KIND, either express or implied.
# See the License for the specific language governing permissions and
# limitations under the License.
"""Vanilla coordinator"""
from collections import OrderedDict
from typing import Any, List, Optional, Type, Union

from adorn.alter.alter import Alter
from adorn.exception.type_check_error import TypeCheckError
from adorn.orchestrator.orchestrator import Orchestrator
from adorn.unit.unit import Unit


class Base(Orchestrator):
    """Vanilla coordinator between different :class:`~adorn.unit.unit.Unit`

    ``Base`` finds the first :class:`~adorn.unit.unit.Unit` specified in
    the list of :class:`~adorn.unit.unit.Unit` that contains a type
    and delegates the action to the given :class:`~adorn.unit.unit.Unit`.
    Therefore, the earlier a :class:`~adorn.unit.unit.Unit` appears in
    ``units`` the higher priority it has in the orchestrator.

    Args:
        units (List[Unit]): a collection of collection of types
        alters (List[Alter]): a collection of ways to dynamically alter
            a configuration

    Attributes:
        unit_dict (OrderedDict[Type, Unit]): a mapping from a type of a
            :class:`~adorn.unit.unit.Unit` to an instance of the
            :class:`~adorn.unit.unit.Unit`
    """

    def __init__(self, units: List[Unit], alters: Optional[List[Alter]] = None) -> None:
        super().__init__()
        self.units = units
        self.alters = alters or []
        self.unit_dict = OrderedDict([(type(i), i) for i in self.units])

    def contains(self, cls: Type) -> bool:
        """Check if ``Type`` is contained in one of the :class:`~adorn.unit.unit.Unit` objects

        Args:
            cls (Type): a type that is being checked to see if
                it exists in one of the :class:`~adorn.unit.unit.Unit`

        Returns:
            bool: if ``True`` a :class:`~adorn.unit.unit.Unit`
                contained in the orchestrator contains
                the provided type
        """  # noqa: B950
        return any(i.contains(cls, self) for i in self.units)

    def get(self, cls: Type) -> Unit:
        """Finds the :class:`~adorn.unit.unit.Unit` associated with `cls`

        .. note::

            The first :class:`~adorn.unit.unit.Unit` will be returned,
            where order is determined by a
            :class:`~adorn.unit.unit.Unit` location in ``units``.

        Args:
            cls (Type): a type that you want the associated ``Unit`` of

        Returns:
            Unit: The collection of types that contains ``cls``

        Raises:
            TypeCheckError: ``cls`` cannot be represented by any of
                the :class:`~adorn.unit.unit.Unit` contained in the
                orchestrator
        """
        for i in self.units:
            if i.contains(cls, self):
                return i
        ru_str = "\n".join([f"{type(i)}" for i in self.units])
        raise TypeCheckError(
            target_cls=cls,
            msg=[
                f"{cls} is not supported by any of the following Unit's:",
                f"{ru_str}",
            ],
        )

    def a_get(self, cls: Type, obj: Any) -> List[Alter]:
        """Finds all the :class:`~adorn.alter.alter.Alter` associated with `cls`

        .. note::

            The list of :class:`~adorn.alter.alter.Alter` will be returned,
            in the order :class:`~adorn.alter.alter.Alter` were passed.

        Args:
            cls (Type): a type that you want to check or convert the ``obj`` into
            obj (Any): the object to be type checked or converted into ``cls``

        Returns:
            List[Alter]: the classes requested to perform a dynamic alteration
                to the ``obj``
        """
        return [i for i in self.alters if i.a_contains(cls, self, obj)]

    def alter_obj(self, cls: Type, obj: Any) -> Union[Any, TypeCheckError]:
        """Potentially perform a dynamic alteration to an object

        Args:
            cls (Type): a type that you want to check or convert the ``obj`` into
            obj (Any): the object to be type checked or converted into ``cls``

        Returns:
            Union[Any, TypeCheckError]: Either an unaltered ``obj`` is returned, when no
                :class`~adorn.alter.alter.Alter` in ``alters`` contains the
                given ``cls`` and ``obj``, an altered ``obj`` is returned when the
                given ``cls`` and ``obj`` is contained by an
                :class`~adorn.alter.alter.Alter` in ``alters``, or a ``TypeCheckError``,
                when there was an issue performing a dynamic alteration to the ``obj``
        """
        alter_list = self.a_get(cls, obj)
        for alter in alter_list:
            obj = alter.alter_obj(cls, self, obj)
        return obj

    def log_instance(self, cls: Type, obj: Any, instance: Any) -> None:
        """Potentially perform a dynamic logging of instantiated instances

        Args:
            cls (Type): a type that you converted the ``obj`` into
            obj (Any): the object converted into ``cls``
            instance (Any): the object created from converting ``obj`` into an
                instance of ``cls``
        """
        alter_list = self.a_get(cls, obj)
        for alter in alter_list:
            obj = alter.log_instance(cls, self, obj, instance)

    def type_check(self, cls: Type, obj: Any) -> Optional[TypeCheckError]:
        """Check if `obj` can be converted to type of `cls`

        First, the code gets the :class:`~adorn.unit.unit.Unit` that
        contains ``cls``.  Then the type checking is delegated to the
        relevant :class:`~adorn.unit.unit.Unit`.

        Args:
            cls (Type): the type ``obj`` would be converted into
            obj (Any): an instance to be converted into type ``cls``

        Returns:
            Optional[TypeCheckError]: if ``None``, ``obj`` can be converted to an
                instance of ``cls``, otherwise an error is returned explaining
                why an ``obj`` cannot be converted to an instance of ``cls``
        """
        obj = self.alter_obj(cls, obj)
        if isinstance(obj, TypeCheckError):
            return obj
        ru = self.get(cls)
        return ru.type_check(cls, self, obj)

    def from_obj(self, cls: Type, obj: Any) -> Any:
        """Generate an instance of type `cls` from `obj`

        First, the code gets the :class:`~adorn.unit.unit.Unit` that
        contains ``cls``.  Then the instantiating is delegated to the
        relevant :class:`~adorn.unit.unit.Unit`.

        Args:
            cls (Type): the type ``obj`` would be converted into
            obj (Any): an instance to be converted into type ``cls``

        Returns:
            Any: an instance of type ``cls``, that was created from ``obj``

        Raises:
            obj: obj is a ``TypeCheckError`` indicating there was an issue
                applying a dynamic alteration to ``obj``
        """
        obj = self.alter_obj(cls, obj)
        if isinstance(obj, TypeCheckError):
            raise obj
        ru = self.get(cls)
        output = ru.from_obj(cls, self, obj)
        self.log_instance(cls, obj, output)
        return output
